/*****************************************************/
/*  textmode.c                                       */
/* -- Sample program demonstrates 80x25 mode from    */
/*    within Windows.                                */
/*****************************************************/

#include <windows.h>
#include <memory.h>
#include "dpmi.h"

/*****************************************************/
/* Prototypes.                                       */
/*****************************************************/
/* Private. */
int              CbGetStateBufferSize(VOID);
VOID             ExitTextMode(HWND);
VOID             EnterTextMode(HWND);
VOID             HandleKey(WPARAM);
VOID             SetCursorXY(char, char);
VOID             SetLpvWCw(VOID FAR *, WORD, UINT);
VOID             VideoInterrupt(RMI *);

/* Exports. */
LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

/* Imports. */
DWORD WINAPI     GlobalDOSAlloc(DWORD);
UINT  WINAPI     GlobalDOSFree(UINT);

/*****************************************************/
/* Constants.                                        */
/*****************************************************/
#define szApp    "Video"
#define dxText   80
#define dyText   25
#define cbText   (dxText * dyText * sizeof(WORD))
#define cbVideoBlock    64  /* Video block in bytes. */

/*****************************************************/
/* Globals.                                          */
/*****************************************************/
/* Imports. */
extern WORD _B000H;

/* Local. */
LPWORD      lrgwText, lrgwSav;
char        x, y;
BOOL        fInTextMode;
BOOL        fWantTextMode;
RECT        rectWindow;
WORD        selStateBuffer; /* Buffer selector. */
WORD        segStateBuffer; /* Real mode segment. */
BYTE        bModeSav;       /* Previous video mode. */

/*****************************************************/
/* Routines.                                         */
/*****************************************************/

int FAR PASCAL
WinMain(HINSTANCE hins, HINSTANCE hinsPrev, LPSTR lsz,
  int wShow)
/*****************************************************/
/* -- hins        : This program's instance.         */
/* -- hinsPrev    : Previous program's instance.     */
/* -- lsz         : Command line I was invoked with. */
/* -- wShow       : ShowWindow command.              */
/*****************************************************/
    {
    MSG     msg;
    HGLOBAL hmemText        = NULL;
    HWND    hwnd;
    int     cbStateBuffer;
    DWORD   lw;

    if ((cbStateBuffer = CbGetStateBufferSize()) == -1)
        goto WinMainExit;

    if (!(lw = GlobalDOSAlloc(cbStateBuffer)))
        goto WinMainExit;

    segStateBuffer = HIWORD(lw);
    selStateBuffer = LOWORD(lw);

    if (hinsPrev == NULL)
        {
        WNDCLASS    wcs;

        /* First instance is responsible for */
        /* registering a class. */
        wcs.style = CS_HREDRAW | CS_VREDRAW;
        wcs.lpfnWndProc = WndProc;
        wcs.cbClsExtra = 0;
        wcs.cbWndExtra = 0;
        wcs.hInstance = hins;
        wcs.hIcon = NULL;
        wcs.hCursor = LoadCursor(NULL, IDC_ARROW);
        wcs.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
        wcs.lpszMenuName = NULL;
        wcs.lpszClassName = szApp;

        if (!RegisterClass(&wcs))
            goto WinMainExit;
        }

    /* Initialize the video memory. */
    if (!(hmemText =
       GlobalAlloc(GMEM_MOVEABLE, cbText)))
        goto WinMainExit;

    lrgwSav = (LPWORD)GlobalLock(hmemText);
    lrgwText =
      (LPWORD)MAKELONG(0x0000, (WORD)&_B000H);
    SetLpvWCw(lrgwSav, 0x0720, dxText * dyText);

    /* Create the main window. */
    hwnd = CreateWindow(szApp, szApp,
      WS_OVERLAPPEDWINDOW,
      CW_USEDEFAULT, CW_USEDEFAULT,
      CW_USEDEFAULT, CW_USEDEFAULT,
      NULL, NULL, hins, NULL);

    if (!hwnd)
        goto WinMainExit;

    ShowWindow(hwnd, wShow);

    while (GetMessage(&msg, NULL, 0, 0))
        {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
        }

WinMainExit:
    if (selStateBuffer)
        GlobalDOSFree(selStateBuffer);
    if (hmemText)
        GlobalFree(hmemText);
    return 0;
    }

VOID
SetCursorXY(char x, char y)
/*****************************************************/
/* -- Set the cursor position.                       */
/*****************************************************/
    {
    RMI rmi;

    rmi.regDX.w = x + (y << 8);
    rmi.regBX.w = 0;
    rmi.regAX.w = 0x0200;
    VideoInterrupt(&rmi);
    }

VOID
EnterTextMode(HWND hwnd)
/*****************************************************/
/* -- Enter 80x25 text mode.                         */
/* -- Resize the window to full screen, to prevent   */
/*    any other window from being drawn (will cause  */
/*    true nastiness when VGA is in text mode!).     */
/* -- hwnd : Main window.                            */
/*****************************************************/
    {
    RMI rmi;

    if (fInTextMode)
        return;

    ShowCursor(FALSE);

    /* Save video state in buffer. */
    rmi.regAX.w = 0x1c01;
    rmi.regBX.w = 0x0000;
    rmi.regCX.w = 0x0007;
    rmi.wES = segStateBuffer;
    VideoInterrupt(&rmi);

    /* Remember current video mode. */
    rmi.regAX.w = 0x0f00;
    VideoInterrupt(&rmi);
    bModeSav = LOBYTE(rmi.regAX.w);

    /* Set 80x25 text mode. */
    rmi.regAX.w = 0x0007;
    VideoInterrupt(&rmi);

    /* Restore previous text. */
    _fmemcpy(lrgwText, lrgwSav, cbText);

    /* Restore cursor position. */
    SetCursorXY(x, y);

    /* Save old window position. */
    GetWindowRect(hwnd, &rectWindow);

    /* Make full-screen and topmost. */
    SetWindowPos(hwnd, HWND_TOPMOST, 0, 0,
      GetSystemMetrics(SM_CXSCREEN),
      GetSystemMetrics(SM_CYSCREEN), SWP_NOACTIVATE);

    fInTextMode = TRUE;
    }

VOID
ExitTextMode(HWND hwnd)
/*****************************************************/
/* -- Re-enter previous VGA mode Windows was in.     */
/* -- hwnd : Main window.                            */
/*****************************************************/
    {
    RMI rmi;

    if (!fInTextMode)
        return;

    /* Restore previous video mode. */
    rmi.regAX.w = bModeSav;
    VideoInterrupt(&rmi);

    /* Restore previous VGA state. */
    rmi.regAX.w = 0x1c02;
    rmi.regBX.w = 0x0000;
    rmi.regCX.w = 0x0007;
    rmi.wES = segStateBuffer;
    VideoInterrupt(&rmi);

    fInTextMode = FALSE;
    ShowCursor(TRUE);   /* Restore cursor. */

    /* Restore old window position. */
    SetWindowPos(hwnd, HWND_NOTOPMOST,
      rectWindow.left, rectWindow.top,
      rectWindow.right - rectWindow.left,
      rectWindow.bottom - rectWindow.top,
      SWP_NOACTIVATE);

    InvalidateRect(NULL, NULL, TRUE);
    }

VOID
HandleKey(WPARAM wKey)
/*****************************************************/
/* -- Handle a key event.                            */
/*****************************************************/
    {
    if (!fWantTextMode || !fInTextMode)
        return;

    switch (wKey)
        {
    default:
        {
        WORD    wVideo;
        int     iw;

        wVideo = 0x0700 | (wKey & 0x00ff);
        iw = x + y * dxText;
        lrgwText[iw] = lrgwSav[iw] = wVideo;
        if (++x >= dxText)
            {
            x = 0;
            if (++y >= dyText)
                y = 0;
            }
        }
        break;

    case VK_UP:
        if (--y < 0)
            y = dyText - 1;
        break;

    case VK_DOWN:
        if (++y >= dyText)
            y = 0;
        break;

    case VK_LEFT:
        if (--x < 0)
            x = dxText - 1;
        break;

    case VK_RIGHT:
        if (++x >= dxText)
            x = 0;
        break;

    case VK_PRIOR:
        y = 0;
        break;

    case VK_NEXT:
        y = dyText - 1;
        break;

    case VK_HOME:
        x = y = 0;
        break;

    case VK_END:
        x = dxText - 1;
        y = dyText - 1;
        break;

    case VK_RETURN:
        x = 0;
        if (++y >= dyText)
            {
            RMI rmi;

            y = dyText - 1;

            /* Scroll the window. */
            rmi.regAX.w = 0x0601;
            rmi.regBX.w = 0x0700;
            rmi.regCX.w = 0;
            rmi.regDX.w =
              dxText - 1 + ((dyText - 1) << 8);
            VideoInterrupt(&rmi);
            }
        break;
        }       /* End switch wKey. */

    SetCursorXY(x, y);
    }

int
CbGetStateBufferSize(VOID)
/*****************************************************/
/* -- Get the size of the buffer required to hold    */
/*    the video state data.                          */
/* -- Return -1 if adapter does not support save/    */
/*    restore video state.                           */
/*****************************************************/
    {
    RMI rmi;

    rmi.regAX.w = 0x1c00;
    rmi.regCX.w = 0x0007;
    VideoInterrupt(&rmi);
    return (rmi.regAX.w & 0x00ff) == 0x001c ?
      rmi.regBX.w * cbVideoBlock : -1;
    }

VOID
VideoInterrupt(RMI * prmi)
/*******************************************************/
/* -- Use the DPI Simulate Real Mode Interrupt to make */
/*    the interrupt call.                              */
/*******************************************************/
    {
    prmi->wSS = 0;
    prmi->wSP = 0;

    _asm    mov ax, 0x0300;
    _asm    mov bx, 0x0010;
    _asm    mov cx, 0x0000;
    _asm    push ds;
    _asm    pop es;
    _asm    mov di, prmi;
    _asm    int 0x31;
    }
